#include	<stdio.h>
#include	<math.h>

/*
	back - back projector

	Carl Crawford
	Purdue University
	W. Lafayette, IN. 47907

	March 6, 1978

 */

/*
	syntax: back [options]

	VARIABLES						DEFAULT
	v:	number of projections				50
	k:	same as v:					50
	n:      number of rays per projection                   100
	d:	input file					for002.dat
	if:	same as d:					for002.dat
	l:	number of lines in -l option			1
	lines:	same as l:					1
	skip:	number of projections to ignore			0
	x:	line in space for -l				0.0
	y:	same as x:					0.0
	t:	initial object rotation				0.0
	theta:	same as t:					0.0
	c:	projection center offset			0.0
	cen:	same as c:					0.0
	gx:	x start for grid				-1.0
	gy:	x start for grid				1.0
	gs:	width of grid					2.0
	chi:	full fan width					45.0
	gamma:	half fan width					22.5
	q:	quadrant of reconstuction			entire
	of:	output file					pic.r
	r:      radius of reconstruction circle                 1.0
	nfl:	number of flashes for fifth generation		100
	lf:	output line in this file			line

	FLAGS							DEFAULT
	-x:	lax's algorithm					off
	-k:	kak's algorithm					off
	-o:	assume circularily symmetrical object		off
	-l:	reconstruct only a line				off
	-c:	shift grid by half step				off
	-5:	fifth generation data				off
	-p:	don't print program status			off
	-a:	equal angular data				off
	-e:	equal space data				on
	-f:     filtered fifth (invokes -5)                     off

*/

int	algo;		/* 0=parallel 1=kak 2=lax 5=fifth */
int	one;		/* 1=use single projection */
int	line;		/* 1=reconstruct only line of picture */
int	views =	50;	/* default number of views */
int	rays =	100;	/* default number of rays */
int	rays1;		/* rays - 1 */
char	*data =	"for002.dat";	/* default input file name */
char    *ofn =	"pic.r";        /* default output file name */
int	npic =	64;	/* size of grid, npic * npic */
int	npic2 =	128;	/* size of grid for line picture */
int	space;		/* 1 = equal angular data */
int	print =	1;	/* 1 = print program info */
float	pic[4096],*p;	/* picture and pointer to it */
float	proj[512];	/* projection data */
float	bb[512];	/* break points for equal angle data */
float	theta,dtheta;	/* angle value for each view */
float	offset;		/* projection offset */
float	grid,grido2;	/* size of grid block */
float	pi =	3.1415926535;
float	convert;	/* conversion between degress and radians */
float	cost,		/* cos(theta) */
	sint,		/* sin(theta) */
	cosd,cosd2,	/* cost * grid */
	sind,sind2;	/* sint * grid */
int	l;		/* integer location of projection value */
int	ll;		/* related to l */
int	fd;		/* input data fd and output fd */
int	ii1,ii2;	/* set if in line routine */
float	yy1s =	0.0;	/* default value of line reconstruction */
float	yy1;		/* related to yy1s */
float	gridx =	-1.0;	/* x starting grid position */
float	gridy =	1.0;	/* y starting grid position */
float	grids =	2.0;	/* size of the grid */
int	lines =	1;	/* number of line sections in the -l option */
int	skip =	0;	/* skip skip projections between reading */
int	psf;		/* 1=make center grid point 0,0 */
int     quad;           /* quadrant reconstruction */
char    *lfn =  "line"; /* line file name */
float   radius = 1.0;	/* radius of reconstruction circle */

/*	variables used only in the kak and lax algorithm */

float	rdc,		/* distance between the center and the screen */
	r1,		/* distance from detectors to source */
	rsc,		/* distance fron source to the center */
	half,		/* set so that delta t in the proj. is one */
	chi =	45.0,	/* fan width */
	t1,             /* like t */
	t1s,		/* related to t1 */
	t1r1,		/* t * r1 */
	t1r1s,		/* modified t1r1 */
	s1,		/* s */
	s1s,		/* modified s1 */
	cosdr1,		/* cosd * r1 */
	sindr1,		/* sind * r1 */
	cosdrsc,	/* cosd / rsc */
	sindrsc,	/* sind / rsc */
	s1rsc,		/* s1 / rsc */
	s1rscs,		/* related to s1rsc */
	tpr;		/* location in the projection */

/*	variables used only in the parallel algorithm */

float	r,		/* location in the projection */
	rs,		/* modified r */
	t,		/* delta in the projections */
	gridt;		/* grid / t */

/*	variables used only for fifth generation */

int	nfl =	100;	/* number of flashes */
float	drho;		/* step between flashes */
float	taus;		/* start of translation */
float	tau;		/* translate position */
float	dtau;		/* step between flashes */
int     filt;           /* 1=filtered fifth */
int	np;		/* related to n */

char	*table[] = {
		"v",
		"k",
		"r",
		"n",
		"d",
		"if",
		"l",
		"lines",
		"skip",
		"x",
		"y",
		"t",
		"theta",
		"c",
		"cen",
		"gx",
		"gy",
		"gs",
		"chi",
		"gamma",
		"q",
		"of",
		"nfl",
		"lf",
		0
	};

char	*algon[] = {
		"parallel",
		"kak's radial",
		"lax's radial",
		"",
		"",
		"fifth generation"
	};

main(argc,argv)
	int	argc;
	char	**argv;
{
	char	cc;
	register	int	i,j;

	while(++argv,--argc){
		if(**argv == '-')while(cc = *++*argv)switch(cc){
			case 'x':	/* use lax's algorithm */
				algo = 2;
				break;
			case 'k':	/* use kak's algorithm */
				algo = 1;
				break;
			case 'o':	/* use single projection */
				one = 1;
				break;
			case 'l':	/* reconstruct only line of picture */
				line = 1;
				break;
			case 'c':	/* make center grid 0,0 */
				psf = 1;
				break;
			case '5':	/* fifth generation data */
				algo = 5;
				break;
			case 'p':	/* don't print program info */
				print = 0;
				break;
			case 'a':	/* equal angular data */
				space = 1;
				break;
			case 'e':	/* equal space data */
				space = 0;
				break;
			case 'f':       /* filtered fifth */
				filt = 1;
				algo = 5;
				break;
			default:
				fprintf(stderr,"bad flag: -%c\n",cc);
				exit(1);
		}else	switch(comm(*argv)){
			case 1:		/* v */
			case 2:		/* k */
				views = atoi(*argv+2);
				break;
			case 3:		/* r */
				radius = atof(*argv + 2);
				break;
			case 4:		/* n */
				rays = atoi(*argv+2);
				break;
			case 5:		/* d */
				data = *argv + 2;
				break;
			case 6:		/* if */
				data = *argv + 3;
				break;
			case 7:		/* l */
				lines = atoi(*argv + 2);
				line = 1;		/* Set the -l option */
				break;
			case 8:		/* lines */
				lines = atoi(*argv + 5);
				break;
			case 9:		/* skip */
				skip = atoi(*argv + 5);
				break;
			case 10:	/* x */
			case 11:	/* y */
				yy1s = atof(*argv + 2);
				break;
			case 12:	/* t */
				theta = atof(*argv + 2);
				break;
			case 13:	/* theta */
				theta = atof(*argv + 6);
				break;
			case 14:	/* c */
				offset = atof(*argv + 2);
				break;
			case 15:	/* cen */
				offset = atof(*argv + 4);
				break;
			case 16:	/* gx */
				gridx = atof(*argv + 3);
				break;
			case 17:	/* gy */
				gridy = atof(*argv + 3);
				break;
			case 18:	/* gs */
				grids = atof(*argv + 3);
				break;
			case 19:	/* chi */
				chi = atof(*argv + 4);
				break;
			case 20:	/* gamma */
				chi = 2.0 * atof(*argv + 6);
				break;
			case 21:        /* q */
				quad = atoi(*argv + 2);
				break;
			case 22:        /* of */
				ofn = *argv + 3;
				break;
			case 23:	/* nfl */
				nfl = atoi(*argv + 4);
				break;
			case 24:        /* lf */
				lfn = *argv + 3;
				line = 1;
				break;
		}
	}
	init();
	if(line)ofn = lfn;
	if(print){
		printf("BACK DATA\n\n");
		printf("n = %d\n",rays);
		printf("k = %d\n",views);
		printf("r = %f\n",radius);
		printf("gx = %f\n",gridx);
		printf("gy = %f\n",gridy);
		printf("gs = %f\n",grids);
		printf("theta = %f\n",theta/convert);
		printf("cen = %f\n",offset);
		printf("%s back projection algorithm\n",algon[algo]);
		printf("input file: %s\n",data);
		printf("output file: %s\n",ofn);
		printf("equal %s data\n",(space)? "angular" : "spaced");
		if(psf)printf("psf mode on\n");
		if(line){
			printf("line reconstruction: ");
			printf("%d line(s)",lines);
			printf(" starting at y = %f\n",yy1s);
		}
		if(one)printf("one projection mode on\n");
		if(skip)printf("skip %d projection(s)\n",skip);
		if(algo){
			printf("chi = %f\n",chi/convert);
			printf("rsc = %f\n",rsc);
			printf("rdc = %f\n",rdc);
			if(filt)printf("np = %d\n",np);
		}
		if(algo == 5){
			printf("nfl = %d\n",nfl);
			printf("%sfiltered version\n",(filt)? "" : "un");
		}
		printf("\n");
	}

	if(one){
		readp();
		for(i=0;i<rays;proj[i++] /= views );
	}
	for(i=0;i<views;i++){
		if(one == 0 && algo != 5)readp();
		switch(algo){

		case 0:
			par();
			break;
		case 1:
			kak();
			break;
		case 2:
			if(space)	laxa();
			else		laxs();
			break;
		case 5:
			tau = taus;
			for(j=0;j<nfl;j++){
				readp();
				if(filt){
					if(space)        fiftha();
					else            fifths();
				}else{
					fifthu();
				}
				theta += drho;
				tau += dtau;
			}
			break;
		}
		theta += dtheta;
	}
	store();
}

readp()
{
	register	int	i;

	if(read(fd,proj,rays*sizeof(float)) != rays*sizeof(float)){
		fprintf(stderr,"insuffcient data\n");
		exit(1);
	}
	if(skip){
		lseek(fd,(long)skip * (rays*sizeof(float)),1);
		for(i=0;i<rays;proj[i++] *= skip);
	}
}

init()
{
	register	int	i;
	float	dchi,chip,tmp;

	if(line == 0){
		npic2 = npic * npic;
		ii1 = 0;
		ii2 = npic;
	}else{
		npic = npic2;
		ii1 = 0;
		ii2 = lines;
		npic2 *= lines;
	}
	switch(quad){      /* set up for quadrant reconstructions */

	case 0:     /* normal */
	    break;
	case 1:
	    grids = 1.0;    gridx = 0.0;    gridy = 1.0;
	    break;
	case 2:
	    grids = 1.0;    gridx = -1.0;   gridy = 1.0;
	    break;
	case 3:
	    grids = 1.0;    gridx = -1.0;   gridy = 0.0;
	    break;
	case 4:
	    grids = 1.0;    gridx = 0.0;    gridy = 0.0;
	    break;
	default:
	    fprintf(stderr,"bad quadrant\n");
	    exit(1);
	}
	for(i=0;i<npic2;pic[i++] = 0.0);
	grid = grids / npic;
	grido2 = grid * 0.5;
	dtheta =  pi / views;
	convert = pi / 180.;
	theta *= convert;
	if(algo) dtheta += dtheta;
	switch(algo){

	case 0:
		t = 2.0 * radius / rays;
		gridt = grid / t;
		break;
	case 1:
	case 2:
		chi *= convert;
		rsc = radius / (sin(chi/2.0)/cos(chi/2.0));
		half = rays/2.0 - 0.5;
		rdc = (half + 0.5) * rsc / radius - rsc;
		r1 = rdc + rsc;
		break;
	case 5:		/* fifth generation */
		if(filt){
			tmp = chi = 450.0 / views * convert;
			rsc = 0.5 * radius / sin(chi * 0.5);
			taus = 1.5 * radius / cos(chi * 0.5);
			dtau = 2.0 * taus / (nfl - 1.0);
			dtheta = chi / 5.0;
			drho = 3.0 * dtheta / nfl ;
			taus = -taus;
			half = rays / 2.0 - 0.5;
			rdc = (half + 0.5) * rsc * cos(chi * 0.5) * 2.0 / radius - rsc;
			r1 = rdc + rsc;
			np = rays;
			rays = (4.0*asin(sin(chi)/sqrt(5.0-cos(chi)*4.0))/chi+1.0)*rays+0.5;
			if((rays-np) & 1)rays++;
			half = rays / 2.0 - 0.5;
			chi = (chi * rays) / np;
			if(!space){
				rays = tan(chi * 0.5) / tan(tmp * 0.5) * np;
				if((rays - np) & 1)rays++;
				half = rays / 2.0 - 0.5;
			}
		}else{
			chi = 450.0 / views * convert;
			rsc = 0.5 * radius / sin(chi * 0.5);
			taus = 1.5 * radius / cos(chi * 0.5);
			dtau = 2.0 * taus / (nfl - 1.0);
			dtheta = chi / 5.0;
			drho = 3.0 * dtheta / nfl;
			taus = -taus;
			half = rays / 2.0 - 0.5;
			rdc = (half + 0.5) * rsc * cos(chi * 0.5) * 2.0 / radius - rsc;
			r1 = rdc + rsc;
		}
		break;
	}

	if(space){	/* set up break points for equal angle data */
		if( algo <= 1){
			fprintf(stderr,"equal angle data illegal\n");
			exit(1);
		}
		dchi = chi / rays;
		chip = -(chi + dchi) * 0.5;
		for(i=0;i<rays;i++){
			bb[i] = tan(chip) * r1;
			chip += dchi;
		}
	}
	rays1 = rays - 1;
	if((fd = open(data,0)) < 0){
		fprintf(stderr,"can't open: %s\n",data);
		exit(1);
	}
	if(line == 0)
		if(psf)	yy1s = gridy;
		else	yy1s = gridy - grido2;
	else
		yy1s += grido2* (lines -1);
}

par()
{
	register	int	i,j;

	cost = cos(theta);
	sint = sin(theta);
	cosd = cost * gridt;
	sind = sint * gridt;

	yy1 = yy1s;
	r =rs = (((grido2 * (1 -psf)) + gridx) * cost + (sint * yy1) +1.0)/t - 0.5 - offset;
	for(p=pic,i=ii1;i<ii2;i++){
		for(j=0;j<npic;j++){
			l = floor(r);
			if(l>=0 && l <rays)*p += proj[l] + (r-l)*(proj[l+1] - proj[l]);
			p++;
			r += cosd;
		}
		yy1 -= grid;
		r = rs -= sind;
	}
}

kak()
{
	register	int	i,j;

	cost = cos(theta);
	sint = sin(theta);
	cosd2 = grido2 * cost;
	sind2 = grido2 * sint;
	cosd = cosd2 + cosd2;
	sind = sind2 + sind2;
	cosdr1 = cosd * r1;
	sindr1 = sind * r1;

	yy1 = yy1s;
	t1r1 = t1r1s = ((gridx * cost) + (yy1 * sint) + cosd2 -cosd) * r1;
	s1 = s1s = (-gridx * sint) + (yy1*cost) + sind - sind2;
	for(i=ii1,p=pic;i<ii2;i++){
		for(j=0;j<npic;j++){
			t1r1 += cosdr1;
			s1 -= sind;
			tpr = t1r1 /(-s1 + rsc);
			r = tpr + half;
			l = floor(r);
			if(l>=0 && l<rays)*p += proj[l] + (l - r) * (proj[l] - proj[l+1]);
			p++;
		}
		yy1 -= grid;
		t1r1 = t1r1s -= sindr1;
		s1 = s1s -= cosd;
	}
}

laxa()
{
	register	int	i,j;

	cost = cos(theta);
	sint = sin(theta);
	cosd2 = grido2 * cost;
	sind2 = grido2 * sint;
	cosd = cosd2 + cosd2;
	sind = sind2 + sind2;
	cosdr1 = cosd * r1;
	sindr1 = sind * r1;
	cosdrsc = cosd / rsc;
	sindrsc = sind / rsc;

	yy1 = yy1s;
	t1 = t1s = ((gridx * cost) + (yy1 * sint) + cosd2 -cosd);
	t1r1 = t1r1s = t1 * r1;
	s1 = s1s = (-gridx * sint) + (yy1*cost) + sind - sind2;
	s1rsc = s1rscs = s1 / rsc;
	for(i=ii1,p=pic;i<ii2;i++){
		for(j=0,ll=0,rs= -1000000.0;j<npic;j++){
			t1 += cosd;
			t1r1 += cosdr1;
			s1 -= sind;
			s1rsc -= sindrsc;
			r = t1r1 /(-s1 + rsc);
			if(r >= rs){
				while(r > bb[ll]){
					if(ll == rays1)break;
					ll++;
				}
				l = ll;
			}else{
				while(r < bb[ll]){
					if(ll == 0)break;
					ll--;
				}
				l = ll + 1;
			}
			if(l>1 && l<rays1)
			   *p += ((r-bb[l-1])*proj[l]+(bb[l]-r)*proj[l-1])/
			    ((bb[l]-bb[l-1])*((rsc-s1) * (rsc - s1) + t1 * t1));
			rs = r;
			p++;
		}
		yy1 -= grid;
		t1r1 = t1r1s -= sindr1;
		t1 = t1s -= sind;
		s1 = s1s -= cosd;
		s1rsc = s1rscs -= cosdrsc;
	}
}

laxs()
{
	register	int	i,j;

	cost = cos(theta);
	sint = sin(theta);
	cosd2 = grido2 * cost;
	sind2 = grido2 * sint;
	cosd = cosd2 + cosd2;
	sind = sind2 + sind2;
	cosdr1 = cosd * r1;
	sindr1 = sind * r1;
	cosdrsc = cosd / rsc;
	sindrsc = sind / rsc;

	yy1 = yy1s;
	t1r1 = t1r1s = ((gridx * cost) + (yy1 * sint) + cosd2 -cosd) * r1;
	s1 = s1s = (-gridx * sint) + (yy1*cost) + sind - sind2;
	s1rsc = s1rscs = s1 / rsc;
	for(i=ii1,p=pic;i<ii2;i++){
		for(j=0;j<npic;j++){
			t1r1 += cosdr1;
			s1 -= sind;
			s1rsc -= sindrsc;
			tpr = t1r1 /(-s1 + rsc);
			r = tpr + half;
			l = floor(r);
			if(l>=0 && l<rays)
			   *p += (proj[l] + (l - r) * (proj[l] - proj[l+1]))/
				((1.0 - s1rsc) * (1.0 - s1rsc));
			p++;
		}
		yy1 -= grid;
		t1r1 = t1r1s -= sindr1;
		s1 = s1s -= cosd;
		s1rsc = s1rscs -= cosdrsc;
	}
}

fifths()
{
	register	int	i,j;

	cost = cos(theta);
	sint = sin(theta);
	cosd2 = grido2 * cost;
	sind2 = grido2 * sint;
	cosd = cosd2 + cosd2;
	sind = sind2 + sind2;
	cosdr1 = cosd * r1;
	sindr1 = sind * r1;
	cosdrsc = cosd / rsc;
	sindrsc = sind / rsc;

	yy1 = yy1s;
	t1r1 = t1r1s = ((gridx * cost) + (yy1 * sint) + cosd2 -cosd - tau) * r1;
	s1 = s1s = (-gridx * sint) + (yy1*cost) + sind - sind2;
	s1rsc = s1rscs = s1 / rsc;
	for(i=ii1,p=pic;i<ii2;i++){
		for(j=0;j<npic;j++){
			t1r1 += cosdr1;
			s1 -= sind;
			s1rsc -= sindrsc;
			tpr = t1r1 /(-s1 + rsc);
			r = tpr + half;
			l = floor(r);
			if(l>=0 && l<rays)*p += (proj[l] + (l - r) * (proj[l] - proj[l+1]))/
			    ((1.0 - s1rsc) * (1.0 - s1rsc));
			p++;
		}
		yy1 -= grid;
		t1r1 = t1r1s -= sindr1;
		s1 = s1s -= cosd;
		s1rsc = s1rscs -= cosdrsc;
	}
}

fifthu()
{
	register	int	i,j;

	cost = cos(theta);
	sint = sin(theta);
	cosd2 = grido2 * cost;
	sind2 = grido2 * sint;
	cosd = cosd2 + cosd2;
	sind = sind2 + sind2;
	cosdr1 = cosd * r1;
	sindr1 = sind * r1;

	yy1 = yy1s;
	t1r1 = t1r1s = ((gridx * cost) + (yy1 * sint) + cosd2 -cosd - tau) * r1;
	s1 = s1s = (-gridx * sint) + (yy1*cost) + sind - sind2;
	for(i=ii1,p=pic;i<ii2;i++){
		for(j=0;j<npic;j++){
			t1r1 += cosdr1;
			s1 -= sind;
			tpr = t1r1 /(-s1 + rsc);
			r = tpr + half;
			l = floor(r);
			if(l>=0 && l<rays)*p += proj[l] + (l - r) * (proj[l] - proj[l+1]);
			p++;
		}
		yy1 -= grid;
		t1r1 = t1r1s -= sindr1;
		s1 = s1s -= cosd;
	}
}

fiftha()
{
	register	int	i,j;

	cost = cos(theta);
	sint = sin(theta);
	cosd2 = grido2 * cost;
	sind2 = grido2 * sint;
	cosd = cosd2 + cosd2;
	sind = sind2 + sind2;
	cosdr1 = cosd * r1;
	sindr1 = sind * r1;
	cosdrsc = cosd / rsc;
	sindrsc = sind / rsc;

	yy1 = yy1s;
	t1 = t1s = ((gridx * cost) + (yy1 * sint) + cosd2 -cosd - tau );
	t1r1 = t1r1s = t1 * r1;
	s1 = s1s = (-gridx * sint) + (yy1*cost) + sind - sind2;
	s1rsc = s1rscs = s1 / rsc;
	for(i=ii1,p=pic;i<ii2;i++){
		for(j=0,ll=0,rs= -1000000.0;j<npic;j++){
			t1 += cosd;
			t1r1 += cosdr1;
			s1 -= sind;
			s1rsc -= sindrsc;
			r = t1r1 /(-s1 + rsc);
			if(r >= rs){
				while(r > bb[ll]){
					if(ll == rays1)break;
					ll++;
				}
				l = ll;
			}else{
				while(r < bb[ll]){
					if(ll == 0)break;
					ll--;
				}
				l = ll + 1;
			}
			if(l>1 && l<rays1)*p += ((r-bb[l-1])*proj[l]+(bb[l]-r)*proj[l-1])/
			   ((bb[l] - bb[l-1])*((rsc - s1) * (rsc - s1) + t1 * t1));
			rs = r;
			p++;
		}
		yy1 -= grid;
		t1r1 = t1r1s -= sindr1;
		t1 = t1s -= sind;
		s1 = s1s -= cosd;
		s1rsc = s1rscs -= cosdrsc;
	}
}

store()
{
	if(line){
		if((fd = creat(ofn,0644)) < 0){
			fprintf(stderr,"can't creat: line\n");
			exit(1);
		}
		write(fd,pic,npic2*sizeof(float));
	}else{
		if((fd=creat(ofn,0644))<0){
			fprintf(stderr,"can't create: %s\n",ofn);
			exit(1);
		}
		write(fd,pic,4096*sizeof(float));
	}
}

comm(s)
	char	*s;
{
	register	int	i,j,r;

	for(i=0;table[i];i++){
		for(j=0;(r=table[i][j]) == s[j] && r;j++);
		if(r == 0 && s[j] == '=' && s[j+1] )return(i+1);
	}
	fprintf(stderr,"bad option: %s\n",s);
	exit(1);
	return(0);
}
