#include	<stdio.h>
#include	<math.h>

/*
 	cen - find center of scanner data
 
	Carl Crawford
 	Purdue University
 	West Lafayette, Indiana 47907
 
 	April 23,1981
 */

/*
	syntax: cen [options]

	VARIABLES:						DEFAULT
	n:	number of detectors (maximum 100)		100
	if:	input file name for ellipse data		for003.dat
	k:	number of projections				100

	FLAGS:							DEFAULT
	-f:	don't flip last projection
	-p:	don't print program status
	-e:	save error file
	-c:	generate command line output format

*/

#define	MD	512	/* max number of detectors */

int	n =	100;	/* number of detectors */
int	n2;		/* n * 3 */
char	*ifn =	"for003.dat";  /* input file */
FILE	*ifd;		/* input file descriptor */
int	k =	100;	/* number of projections */
float	pr1[3*MD];	/* zero degree projection */
float	pr2[MD];	/* 180 degree projection */
float	merr[2*MD];	/* error array */
int	print;		/* 1=don't print program info */
int	flip;		/* 1=don't flip last projection */
float	diff;		/* tmp difference value */
int	serr;		/* 1=save error vector in "err" */
int	com;		/* 1=use commandline format */

char	*table[] = {
		"n",
		"k",
		"if",
		0
	};

main(argc,argv)
	int	argc;
	char	**argv;
{
	register	i,j;
	char	cc;

	while(argv++,--argc){
		if(**argv == '-')while(cc = *++*argv)switch(cc){
			case 'p':	/* print flag */
				print = 1;
				break;
			case 's':	/* don't flop */
			case 'f':
				flip = 1;
				break;
			case 'e':	/* save error vector */
				serr = 1;
				break;
			case 'c':	/* command line format */
				com = 1;
				break;
			default:
				fprintf(stderr,"bad flag: -%c\n",cc);
				exit(1);
		}else	switch(comm(*argv)){
			case 1:		/* n */
				n = atoi(*argv + 2);
				break;
			case 3:		/* if */
				ifn = *argv + 3;
				break;
			case 2:		/* k */
				k = atoi(*argv + 2);
				break;
		}
	}

	if(!print && !com){
		printf("CEN DATA\n\n");
		printf("n: %d\n",n);
		printf("k: %d\n",k);
		printf("if: %s\n",ifn);
		if(flip)printf("no swap\n");
		if(serr)printf("error vector saved in 'err'\n");
	}

	if(n > MD || n < 1)err("bad detector count","");
	if((ifd = fopen(ifn,"r")) == NULL)err("can't open: ",ifn);
	n2 = n * 2;

	for(i=0;i<(n2+n);pr1[i++] = 0.0);
	fread(pr1+n,sizeof(*pr1),n,ifd); 
	fseek(ifd,(long)sizeof(*pr1)*(k-2)*n,1);
	if(fread(pr2,sizeof(*pr2),n,ifd) != n)err("unexpected EOF","");
	slant(pr1+n);
	slant(pr2);
	if(!flip)swap(pr2);
	for(i=0;i<n2;i++){
		for(j=0,merr[i]=0.0;j<n;j++){
			diff = pr2[j] - pr1[i+j];
			merr[i] += diff * diff;
		}
	}
	if(serr)fwrite(merr,sizeof(*merr),n2,fopen("err","w"));

	diff = HUGE;
	for(j=i=0;i<n2;i++)if(merr[i] < diff){
		diff = merr[i];
		j = i;
	}
	if(!com)printf("cen=");
	printf("%f\n",((float)(n-j))/2.0);
}


comm(s)
	char	*s;
{
	register	int	i,j,r;

	for(i=0;table[i];i++){
		for(j=0;(r=table[i][j]) == s[j] && r;j++);
		if(r == 0 && s[j] == '=' && s[j+1] )return(i+1);
	}
	fprintf(stderr,"bad option: %s\n",s);
	exit(1);
	return(0);
}

err(s1,s2)
	char    *s1,*s2;
{
	fprintf(stderr,"%s%s\n",s1,s2);
	exit(1);
}

slant(proj)
	float	*proj;
{
	float   ave,avb,*p;
	int     i;

	avb = (proj[0] + proj[1]) * 0.5;
	ave = (proj[n-2] + proj[n-1]) * 0.5;
	ave = (ave - avb) / n;
	for(i=0,p=proj;i<n;i++){
		*p++ -= avb;
		avb += ave;
	}
}

swap(proj)
	float	*proj;
{
	int	i,n2;
	float	tmp,*q,*p;

	n2 = n>>1;
	for(p=proj,q=proj+n-1,i=0;i<n2;i++){
		tmp = *p;
		*p++ = *q;
		*q-- = tmp;
	}
}
