#include	<stdio.h>
#include	<math.h>

/*
	hf - homomorphic filt UCT waveforms

	Carl Crawford
	Purdue University
	W. Lafayette, IN 47907

	Feb. 13, 1981
*/

#define	L	128	/* half length of spectrum */
#define L2      (L + L)
#define	M	8	/* log base 2 of 2L */
#define MN	256	/* maximum n */

int	n = 100;	/* number of samples per projection */
int	k = 1;		/* number of projections */
float	gam = 1.0;	/* gamma */
float	gamp = 1.0;	/* gamma prime */

char    *ifn = "udata";  /* input file name */
char	*ofn = "for003.dat"; /* output file name */
char    *sfn = "spectrum"; /* save magnitude of spectrums */
int     spec;           /* 1=save spectrum */
float	proj[MN];	/* projection */
float   a[L2];          /* real part of fft array */
float   b[L2];          /* imaginary part of fft array */
float   aa[L2];		/* real part of fft array */
float   bb[L2];		/* imaginary part of fft array */
int     tsig[L2];       /* time signal */
FILE    *ifd,*ofd,*sfd,*cfd,*ssfd; /* file descriptors */
int	ssspec;		/* save final spectrum */
float   m0,m1;          /* moments */
int     nofilt;         /* 1=no filter */
int     time;           /* 1=data in time domain */
float   tmp;            /* temp var */
int     b1 = 10;	/* band for division */
int     b2 = 50;        /* band for division */
float	sigma = 3.0;	/* square of standard deviation */
int     div;            /* 1=spectral division */
float   c0,c1;          /* constants for spectral division */
float   df;             /* frequency division */
float   dt = .05;       /* time division */
float   fg[L];          /* f to the power gamma */
float   k11,k12,k22;    /* constants for div */
int     band = 20;      /* part of cepstrum kept */
int     print = 1;      /* 1 = print info */
int     swp;            /* indicator for swap subroutine */
int	raw;		/* 1= raw mode */
int	sw;		/* 1=don't swap */
int	cep;		/* 1=save cepstrum */
char	*cfn= "cepstrum";	/* cepstrum file name */
int	logt;		/* 1=log taken of power spectrum */

char    *table[] = {
		"n",
		"of",
		"if",
		"k",
		"sf",
		"gammap",
		"sigma",
		"b1",
		"b2",
		"band",
		"gamma",
		"dt",
		"cf",
		0
	};

main(argc,argv)
	int	argc;
	char	**argv;
{
	char cc;
	register ii,j,i;
	float fv();

	while(argv++ , --argc){
		if(argv[0][0] == '-')while(cc = *++*argv)switch(cc){

			case 'p':	/* print flag */
				print = 0;
				break;
			case 'd':       /* spectral division */
				div = 1;
				break;
			case 'n':       /* no filtering */
				nofilt = 1;
				break;
			case 't':       /* data in time domain */
				time = 1;
				break;
			case 's':       /* save spectrum */
				spec = 1;
				break;
			case 'S':	/* save final spectrum */
				ssspec = 1;
				break;
			case 'c':	/* save cepstrum */
				cep = 1;
				break;
			case 'r':	/* raw mode */
				raw = 1;
				break;
			case 'w':	/* don't swap */
				sw = 1;
				break;
			default:
				fprintf(stderr,"bad flag: -%c\n",cc);
				exit(1);
			}
		else
			switch(i = comm(*argv)){

			case 1:
				n = atoi(*argv + 2);
				break;
			case 2:
				ofn = *argv + 3;
				break;
			case 3:
				ifn = *argv + 3;
				break;
			case 4:
				k = atoi(*argv + 2);
				break;
			case 5:
				sfn = *argv + 3;
				spec = 1;
				break;
			case 6:
				gamp = atof(*argv + 7);
				break;
			case 7:
				sigma = atof(*argv + 6);
				break;
			case 8:
				b1 = atoi(*argv + 3);
				break;
			case 9:
				b2 = atoi(*argv + 3);
				break;
			case 10:
				band = atoi(*argv + 5);
				break;
			case 11:
				gam = atof(*argv + 6);
				break;
			case 12:
				dt = atof(*argv + 3);
				break;
			case 13:
				cfn = *argv +3;
				break;
		}
	}

	if(n <= 0 || n > MN)err("illegal n","");
	if(ifn[0] == '-' && !ifn[1]){
		ifd = stdin;
	}else{
		if((ifd = fopen(ifn,"r")) == NULL)err("can't open: ",ifn);
	}
	if((ofd = fopen(ofn,"w")) == NULL)err("can't create: ",ofn);
	if(spec)if((sfd = fopen(sfn,"w")) == NULL)err("can't create: ",sfn);
	if(cep)if((cfd = fopen(cfn,"w")) == NULL)err("can't create: ",cfn);
	if(ssspec)if((ssfd = fopen("fspec","w")) == NULL)err("can't create: ","fspec");

	if(print){
		printf("HF DATA\n\n");
		printf("n: %d\n",n);
		printf("k: %d\n",k);
		printf("sigma: %f\n",sigma);
		printf("gamma: %f\n",gam);
		printf("gammap: %f\n",gamp);
		printf("dt: %f\n",dt);
		if(div){
			printf("b1: %d\n",b1);
			printf("b2: %d\n",b2);
		}
		printf("band: %d\n",band);
		printf("input file: %s\n",(ifd == stdin)? "stdin" : ifn);
		printf("output file: %s\n",ofn);
		printf("%s technique\n",div?"division" : "fs");
		if(time)printf("time domain data\n");
		if(nofilt)printf("no homomorphic filtering\n");
		if(spec)printf("spectrum saved in: %s\n",sfn);
		if(cep)printf("cepstrum saved in: %s\n",cfn);
		if(ssspec)printf("final spectrum saved in: fspec\n");
		if(raw)printf("raw mode\n");
		if(sw)printf("no projection swapping\n");
	}


	df = 1.0 / (L2 * dt);
	if(!div){
		b1 = 3;
		b2 = L - 1;
	}
	k11 = k12 = k22 = 0.0;
	for(i=b1;i<=b2;i++){
		fg[i] = pow(i*df,gam);
		k11 += 1;
		k12 += fg[i];
		k22 += fg[i] * fg[i];
	}
	c0 = k11 / (k11 * k22 - k12 * k12) / 2.0;
	c1 = k12 / (k11 * k22 - k12 * k12) / 2.0;

	for(ii=0;ii<k;ii++){	/* loop through projections */
		for(j=0;j<n;j += 2){	/* loop through samples */
			if(time){
				ffread(tsig,sizeof(*tsig),L2,ifd);
				for(i=0;i<L2;i++)aa[i] = tsig[i];
				ffread(tsig,sizeof(*tsig),L2,ifd);
				for(i=0;i<L2;i++)bb[i] = tsig[i];
				fft(aa,bb,M,0);
				for(i=0;i<L2;i++){	/* uncouple FFT */
					m0 = aa[i] + aa[L2 - i];
					m1 = bb[i] - bb[L2 - i];
					a[i] = m0 * m0 + m1 * m1;
					m0 = bb[i] + bb[L2 - i];
					m1 = aa[i] - aa[L2 - i];
					b[i] = m0 * m0 + m1 * m1;
				}
			}else{
				ffread(a,sizeof(float),L,ifd);
				ffread(b,sizeof(float),L,ifd);
			}
			a[0] = a[1];	/* get rid of dc */
			b[0] = b[1];

			if(spec){
				fwrite(a,sizeof(*a),L,sfd);
				fwrite(b,sizeof(*b),L,sfd);
			}
			if(!nofilt){
				for(i=0;i<L;i++){
					a[i] = log(a[i]);
					b[i] = log(b[i]);
				}
				for(i=L+1;i<L2;i++){
					a[i] = a[L2 - i];
					b[i] = b[L2 - i];
				}
				fft(a,b,M,0);
				if(cep){
					fwrite(a,sizeof(*a),L,cfd);
					fwrite(b,sizeof(*b),L,cfd);
				}
				for(i=band;i<=(L2-band);i++)b[i] = a[i] = 0.0;
				fft(a,b,M,1);
				logt = 1;
				if(ssspec){
					fwrite(a,sizeof(*a),L,ssfd);
					fwrite(b,sizeof(*b),L,ssfd);
				}
			}
			proj[j] = fv(a);
			proj[j+1] = fv(b);
		}
		if(!raw){
			slant();
			if(!sw)swap();
		}
		fwrite(proj,sizeof(float),n,ofd);
	}
}

err(s1,s2)
	char    *s1,*s2;
{
	fprintf(stderr,"%s%s\n",s1,s2);
	exit(1);
}

comm(s)
	char	*s;
{
	register	int	i,j,r;

	for(i=0;table[i];i++){
		for(j=0;(r=table[i][j]) == s[j] && r;j++);
		if(r == 0 && s[j] == '=' && s[j+1] )return(i+1);
	}
	fprintf(stderr,"bad option: %s\n",s);
	exit(1);
}

slant()
{
	float   ave,avb,*p;
	register int     i;

	if(!div){
		for(i=1;i<n;i++){
			proj[i] = (proj[0] - proj[i]) / (pow(proj[i],gamp-1.0) * 2.0 * sigma * df * gam);
		}
		proj[0] = 0.0;
	}

	avb = (proj[0] + proj[1]) * 0.5;
	ave = (proj[n-2] + proj[n-1]) * 0.5;
	ave = (ave - avb) / n;
	for(i=0,p=proj;i<n;i++){
		*p++ -= avb;
		avb += ave;
	}
}

swap()
{
	int	i,n2;
	float   tmp,*q,*p;

	if(!swp){
		swp = 1;
		return;
	}else{
		swp = 0;
		n2 = n>>1;
		for(p=proj,q=proj+n-1,i=0;i<n2;i++){
			tmp = *p;
			*p++ = *q;
			*q-- = tmp;
		}
	}
}

ffread(cp,size,n,fd)
	char	*cp;
	int	size,n;
	FILE *fd;
{
	if(fread(cp,size,n,fd) != n)err("unexpected EOF","");
}
float fv(a)
	float	*a;
{
	register int i;

	m1 = m0 = 0.0;
	for(i=b1;i<=b2;i++){
		if(div){
			tmp = (logt)? a[i] : log(a[i]);
			m0 += fg[i] * tmp;
			m1 += tmp;
		}else{
			tmp = (logt)? exp(a[i]) : a[i];
			m0 += tmp;
			m1 += tmp * i;
		}
	}
	if(div){
		return(-c0 * m0 + c1 * m1);
	}else{
		return(m1 / m0);
	}
}
